@@ -9,19 +9,26 @@ __all__ = [
     "Executor",
     "JsonFormat",
     "JsonFormatIterator",
+    "SqliteDictFormat",
     "Step",
     "StepCache",
     "LocalStepCache",
+    "Workspace",
+    "MemoryWorkspace",
+    "LocalWorkspace",
 ]
 
-from tango.executor import Executor, SimpleExecutor
+from tango.executor import Executor
 from tango.format import (
     DillFormat,
     DillFormatIterator,
     Format,
     JsonFormat,
     JsonFormatIterator,
+    SqliteDictFormat,
 )
+from tango.local_workspace import LocalWorkspace
 from tango.step import Step
 from tango.step_cache import LocalStepCache, StepCache
-from tango.step_graph import StepGraph, StepStub
+from tango.step_graph import StepGraph
+from tango.workspace import MemoryWorkspace, Workspace
