@@ -1,10 +1,10 @@
 import os
 import warnings
+from typing import Optional
 
 from filelock import AcquireReturnProxy
 from filelock import FileLock as _FileLock
 from filelock import Timeout
-from overrides import overrides
 
 from .aliases import PathOrStr
 
@@ -24,10 +24,9 @@ class FileLock(_FileLock):  # type: ignore[valid-type,misc]
         super().__init__(str(lock_file), timeout=timeout)
         self._read_only_ok = read_only_ok
 
-    @overrides
-    def acquire(self, timeout=None, poll_intervall=0.05) -> AcquireReturnProxy:
+    def acquire(self, timeout=None, poll_interval=0.05) -> AcquireReturnProxy:
         try:
-            return super().acquire(timeout=timeout, poll_intervall=poll_intervall)
+            return super().acquire(timeout=timeout, poll_interval=poll_interval)
         except OSError as err:
             # OSError could be a lot of different things, but what we're looking
             # for in particular are permission errors, such as:
@@ -47,7 +46,7 @@ class FileLock(_FileLock):  # type: ignore[valid-type,misc]
             else:
                 raise
 
-    def acquire_with_updates(self, desc: str = "acquiring lock"):
+    def acquire_with_updates(self, desc: Optional[str] = None):
         """
         Same as :meth:`acquire()`, except that when the lock cannot be immediately acquired,
         it will keep trying and print status updates as it goes.
@@ -59,6 +58,9 @@ class FileLock(_FileLock):  # type: ignore[valid-type,misc]
 
         from .tqdm import Tqdm
 
+        if desc is None:
+            desc = f"acquiring lock at {self._lock_file}"
+
         progress = Tqdm.tqdm(desc=desc, bar_format="{desc} [{elapsed}]")
         while True:
             progress.update()
