@@ -20,6 +20,11 @@ from typing import (
     get_type_hints,
 )
 
+from tango.common._det_hash import CustomDetHash
+from tango.common.exceptions import ConfigurationError
+from tango.common.lazy import Lazy
+from tango.common.params import Params
+
 try:
     # For PEP 604 support (python >= 3.10)
     from types import UnionType  # type: ignore[attr-defined]
@@ -29,11 +34,6 @@ except ImportError:
         pass
 
 
-from ._det_hash import CustomDetHash
-from .exceptions import ConfigurationError
-from .lazy import Lazy
-from .params import Params
-
 logger = logging.getLogger(__name__)
 
 T = TypeVar("T", bound="FromParams")
@@ -78,21 +78,6 @@ def takes_kwargs(obj) -> bool:
     )
 
 
-def can_construct_from_params(type_: Type) -> bool:
-    if type_ in [str, int, float, bool]:
-        return True
-    origin = getattr(type_, "__origin__", None)
-    if origin == Lazy:
-        return True
-    elif origin:
-        if hasattr(type_, "from_params"):
-            return True
-        args = getattr(type_, "__args__")
-        return all(can_construct_from_params(arg) for arg in args)
-
-    return hasattr(type_, "from_params")
-
-
 def is_base_registrable(cls) -> bool:
     """
     Checks whether this is a class that directly inherits from Registrable, or is a subclass of such
@@ -156,7 +141,38 @@ def infer_method_params(
             # `from __future__ import annotation`, the annotation will be a str,
             # so we need to resolve it using `get_type_hints` from the typing module.
             # See https://www.python.org/dev/peps/pep-0563/ for more info.
-            parameters[param_name] = param.replace(annotation=get_type_hints(method)[param_name])
+            try:
+                parameters[param_name] = param.replace(
+                    annotation=get_type_hints(method)[param_name]
+                )
+            except TypeError as e:
+                if "'type' object is not subscriptable" in str(e):
+                    # This can happen when someone uses a type hint like `dict[str, str]`
+                    # instead of `Dict[str, str]`.
+                    err_msg = (
+                        f"Failed to parse the type annotation `{param.annotation}` "
+                        f"from `{cls.__qualname__}.{method.__name__}()`."
+                    )
+
+                    if "[" in param.annotation:
+                        # Check if there is an equivalent generic in the `typing` module.
+                        import typing
+
+                        type_, *_ = param.annotation.split("[", 1)
+                        for possible_typing_equivalent in {type_, type_.title()}:
+                            if hasattr(typing, possible_typing_equivalent):
+                                err_msg += (
+                                    f" Try using `{possible_typing_equivalent}` "
+                                    "from the `typing` module instead."
+                                )
+                                break
+
+                    new_e = TypeError(err_msg)
+                    new_e.__cause__ = e
+                    new_e.__cause__ = e
+                    raise new_e
+                else:
+                    raise
 
     if var_positional_key:
         del parameters[var_positional_key]
@@ -168,10 +184,13 @@ def infer_method_params(
     # first superclass, and so on. We take the first superclass we find that inherits from
     # FromParams.
     super_class = None
-    for super_class_candidate in cls.mro()[1:]:
-        if issubclass(super_class_candidate, FromParams):
-            super_class = super_class_candidate
-            break
+    # We have to be a little careful here because in some cases we might not have an
+    # actual class. Instead we might just have a function that returns a class instance.
+    if hasattr(cls, "mro"):
+        for super_class_candidate in cls.mro()[1:]:
+            if issubclass(super_class_candidate, FromParams):
+                super_class = super_class_candidate
+                break
     if super_class:
         super_parameters = infer_params(super_class)
     else:
@@ -181,7 +200,10 @@ def infer_method_params(
 
 
 def create_kwargs(
-    constructor: Callable[..., T], cls: Type[T], params: Params, **extras
+    constructor: Callable[..., T],
+    cls: Type[T],
+    params: Params,
+    extras: Optional[Dict[str, Any]] = None,
 ) -> Dict[str, Any]:
     """
     Given some class, a ``Params`` object, and potentially other keyword arguments,
@@ -226,7 +248,7 @@ def create_kwargs(
 
         explicitly_set = param_name in params
         constructed_arg = pop_and_construct_arg(
-            cls.__name__, param_name, annotation, param.default, params, **extras
+            cls.__name__, param_name, annotation, param.default, params, extras or {}
         )
 
         # If the param wasn't explicitly set in `params` and we just ended up constructing
@@ -241,8 +263,9 @@ def create_kwargs(
     if accepts_kwargs:
         for key in list(params):
             kwargs[key] = params.pop(key, keep_as_dict=True)
-        for key, value in extras.items():
-            kwargs[key] = value
+        if extras:
+            for key, value in extras.items():
+                kwargs[key] = value
     params.assert_empty(cls.__name__)
     return kwargs
 
@@ -276,7 +299,12 @@ def create_extras(cls: Type[T], extras: Dict[str, Any]) -> Dict[str, Any]:
 
 
 def pop_and_construct_arg(
-    class_name: str, argument_name: str, annotation: Type, default: Any, params: Params, **extras
+    class_name: str,
+    argument_name: str,
+    annotation: Type,
+    default: Any,
+    params: Params,
+    extras: Dict[str, Any],
 ) -> Any:
     """
     Does the work of actually constructing an individual argument for
@@ -318,29 +346,79 @@ def pop_and_construct_arg(
     return construct_arg(class_name, name, popped_params, annotation, default, **extras)
 
 
+def _params_contain_step(o: Any) -> bool:
+    from tango.step import Step
+
+    if isinstance(o, Step):
+        return True
+    elif isinstance(o, str):
+        return False  # Confusingly, str is an Iterable of itself, resulting in infinite recursion.
+    elif isinstance(o, Params):
+        return _params_contain_step(o.as_dict(quiet=True))
+    elif isinstance(o, dict):
+        if set(o.keys()) == {"type", "ref"} and o["type"] == "ref":
+            return True
+        else:
+            return _params_contain_step(o.values())
+    elif isinstance(o, Iterable):
+        return any(_params_contain_step(p) for p in o)
+    else:
+        return False
+
+
 def construct_arg(
     class_name: str,
     argument_name: str,
     popped_params: Params,
     annotation: Type,
     default: Any,
+    try_from_step: bool = True,
     **extras,
 ) -> Any:
     """
     The first two parameters here are only used for logging if we encounter an error.
     """
-    from tango.step import Step
+    # If we have the default, we're already done :)
+    if popped_params is default:
+        return popped_params
+
+    from tango.step import Step, WithUnresolvedSteps
 
     origin = getattr(annotation, "__origin__", None)
     args = getattr(annotation, "__args__", [])
 
+    # Try to guess if `popped_params` might be a step, come from a step, or contain a step.
+    could_be_step = try_from_step and (
+        origin == Step
+        or isinstance(popped_params, Step)
+        or _params_contain_step(popped_params)
+        or (isinstance(popped_params, (dict, Params)) and popped_params.get("type") == "ref")
+    )
+    if could_be_step:
+        # If we think it might be a step, we try parsing as a step _first_.
+        # Parsing as a non-step always succeeds, because it will fall back to returning a dict.
+        # So we can't try parsing as a non-step first.
+        backup_params = deepcopy(popped_params)
+        try:
+            return construct_arg(
+                class_name,
+                argument_name,
+                popped_params,
+                Step[annotation],  # type: ignore
+                default,
+                try_from_step=False,
+                **extras,
+            )
+        except (ValueError, TypeError, ConfigurationError, AttributeError, IndexError):
+            popped_params = backup_params
+
     # The parameter is optional if its default value is not the "no default" sentinel.
     optional = default != _NO_DEFAULT
 
-    if inspect.isclass(annotation) and issubclass(annotation, FromParams):
-        if popped_params is default:
-            return default
-        elif origin is None and isinstance(popped_params, annotation):
+    if (inspect.isclass(annotation) and issubclass(annotation, FromParams)) or (
+        inspect.isclass(origin) and issubclass(origin, FromParams)
+    ):
+        if origin is None and isinstance(popped_params, annotation):
             return popped_params
         elif popped_params is not None:
             subextras = create_extras(annotation, extras)
@@ -356,13 +434,20 @@ def construct_arg(
                     popped_params = Params({"type": popped_params})
             elif isinstance(popped_params, dict):
                 popped_params = Params(popped_params)
-            elif not isinstance(popped_params, Params):
+            elif not isinstance(popped_params, (Params, Step)):
                 raise TypeError(
                     f"Expected a `Params` object, found `{popped_params}` instead while constructing "
                     f"parameter '{argument_name}' for `{class_name}`"
                 )
 
-            result = annotation.from_params(popped_params, **subextras)
+            result: Union[FromParams, WithUnresolvedSteps]
+            if isinstance(popped_params, Step):
+                result = popped_params
+            else:
+                if origin != Step and _params_contain_step(popped_params):
+                    result = WithUnresolvedSteps(annotation.from_params, popped_params)
+                else:
+                    result = annotation.from_params(popped_params, **subextras)
 
             if isinstance(result, Step):
                 expected_return_type = args[0]
@@ -396,7 +481,10 @@ def construct_arg(
         if type(popped_params) in {int, bool}:
             return annotation(popped_params)
         else:
-            raise TypeError(f"Expected {argument_name} to be a {annotation.__name__}.")
+            raise TypeError(
+                f"Expected {argument_name} to be {annotation.__name__}, "
+                f"found {popped_params} ({type(popped_params)})."
+            )
     elif annotation == str:
         # Strings are special because we allow casting from Path to str.
         if type(popped_params) == str or isinstance(popped_params, Path):
@@ -423,16 +511,13 @@ def construct_arg(
     # This is special logic for handling types like Dict[str, TokenIndexer],
     # List[TokenIndexer], Tuple[TokenIndexer, Tokenizer], and Set[TokenIndexer],
     # which it creates by instantiating each value from_params and returning the resulting structure.
-    elif (
-        origin in {collections.abc.Mapping, Mapping, Dict, dict}
-        and len(args) == 2
-        and can_construct_from_params(args[-1])
-    ):
+    elif origin in {collections.abc.Mapping, Mapping, Dict, dict} and len(args) == 2:
         value_cls = annotation.__args__[-1]
         value_dict = {}
         if not isinstance(popped_params, Mapping):
             raise TypeError(
-                f"Expected {argument_name} to be a Mapping (probably a dict or a Params object)."
+                f"Expected {argument_name} to be a Mapping (probably a dict or a Params object) "
+                f"found {popped_params} ({type(popped_params)})."
             )
 
         for key, value_params in popped_params.items():
@@ -447,7 +532,7 @@ def construct_arg(
 
         return value_dict
 
-    elif origin in (Tuple, tuple) and all(can_construct_from_params(arg) for arg in args):
+    elif origin in (Tuple, tuple):
         value_list = []
 
         for i, (value_cls, value_params) in enumerate(zip(annotation.__args__, popped_params)):
@@ -463,7 +548,7 @@ def construct_arg(
 
         return tuple(value_list)
 
-    elif origin in (Set, set) and len(args) == 1 and can_construct_from_params(args[0]):
+    elif origin in (Set, set) and len(args) == 1:
         value_cls = annotation.__args__[0]
 
         value_set = set()
@@ -513,20 +598,13 @@ def construct_arg(
         config_error.__cause__ = error_chain
         raise config_error
     elif origin == Lazy:
-        if popped_params is default:
-            return default
-
         value_cls = args[0]
         return Lazy(value_cls, params=deepcopy(popped_params))  # type: ignore
 
     # For any other kind of iterable, we will just assume that a list is good enough, and treat
     # it the same as List. This condition needs to be at the end, so we don't catch other kinds
     # of Iterables with this branch.
-    elif (
-        origin in {collections.abc.Iterable, Iterable, List, list}
-        and len(args) == 1
-        and can_construct_from_params(args[0])
-    ):
+    elif origin in {collections.abc.Iterable, Iterable, List, list} and len(args) == 1:
         value_cls = annotation.__args__[0]
 
         value_list = []
@@ -544,12 +622,22 @@ def construct_arg(
 
         return value_list
 
-    elif inspect.isclass(annotation) and isinstance(popped_params, Params):
-        subextras = create_extras(annotation, extras)
-        constructor_to_inspect = annotation.__init__
-        constructor_to_call = annotation
-        kwargs = create_kwargs(constructor_to_inspect, annotation, popped_params, **subextras)
-        return constructor_to_call(**kwargs)  # type: ignore
+    elif (inspect.isclass(annotation) or inspect.isclass(origin)) and isinstance(
+        popped_params, Params
+    ):
+        # Constructing arbitrary classes from params
+        arbitrary_class = origin or annotation
+        subextras = create_extras(arbitrary_class, extras)
+        constructor_to_inspect = arbitrary_class.__init__
+        constructor_to_call = arbitrary_class
+        params_contain_step = _params_contain_step(popped_params)
+        kwargs = create_kwargs(constructor_to_inspect, arbitrary_class, popped_params, subextras)
+        from tango.step import WithUnresolvedSteps
+
+        if origin != Step and params_contain_step:
+            return WithUnresolvedSteps(constructor_to_call, *[], **kwargs)
+        else:
+            return constructor_to_call(**kwargs)  # type: ignore
 
     else:
         # Pass it on as is and hope for the best.   ¯\_(ツ)_/¯
@@ -631,6 +719,7 @@ class FromParams(CustomDetHash):
             if "type" in params and params["type"] not in as_registrable.list_available():
                 as_registrable.search_modules(params["type"])
 
+            # Resolve the subclass and constructor.
             if is_base_registrable(cls) or "type" in params:
                 default_to_first_choice = as_registrable.default_implementation is not None
                 choice = params.pop_choice(
@@ -638,19 +727,37 @@ class FromParams(CustomDetHash):
                     choices=as_registrable.list_available(),
                     default_to_first_choice=default_to_first_choice,
                 )
-                subclass, constructor_name = as_registrable.resolve_class_name(choice)
+                # We allow users to register methods and functions, not just classes.
+                # So we have to handle both here.
+                subclass_or_factory_func, constructor_name = as_registrable.resolve_class_name(
+                    choice
+                )
+                if inspect.isclass(subclass_or_factory_func):
+                    # We have an actual class.
+                    subclass = subclass_or_factory_func
+                    if constructor_name is not None:
+                        constructor_to_inspect = cast(
+                            Callable[..., T], getattr(subclass, constructor_name)
+                        )
+                        constructor_to_call = constructor_to_inspect
+                    else:
+                        constructor_to_inspect = subclass.__init__
+                        constructor_to_call = subclass
+                else:
+                    # We have a function that returns an instance of the class.
+                    factory_func = cast(Callable[..., T], subclass_or_factory_func)
+                    return_type = inspect.signature(factory_func).return_annotation
+                    if return_type == inspect.Signature.empty:
+                        subclass = cls
+                    else:
+                        subclass = return_type
+                    constructor_to_inspect = factory_func
+                    constructor_to_call = factory_func
             else:
                 # Must be trying to instantiate the given class directly.
                 subclass = cls
-                constructor_name = None
-
-            # See the docstring for an explanation of what's going on here.
-            if not constructor_name:
-                constructor_to_inspect = subclass.__init__
-                constructor_to_call = subclass  # type: ignore
-            else:
-                constructor_to_inspect = cast(Callable[..., T], getattr(subclass, constructor_name))
-                constructor_to_call = constructor_to_inspect
+                constructor_to_inspect = cls.__init__
+                constructor_to_call = cast(Callable[..., T], cls)
 
             if hasattr(subclass, "from_params"):
                 # We want to call subclass.from_params.
@@ -670,7 +777,8 @@ class FromParams(CustomDetHash):
                 # instead of adding a `from_params` method for them somehow.  We just trust that
                 # you've done the right thing in passing your parameters, and nothing else needs to
                 # be recursively constructed.
-                return constructor_to_call(**params, **create_extras(constructor_to_call, extras))  # type: ignore
+                kwargs = create_kwargs(constructor_to_inspect, subclass, params, extras)  # type: ignore
+                return constructor_to_call(**kwargs)  # type: ignore
         else:
             # This is not a base class, so convert our params and extras into a dict of kwargs.
 
@@ -684,12 +792,12 @@ class FromParams(CustomDetHash):
                 # This class does not have an explicit constructor, so don't give it any kwargs.
                 # Without this logic, create_kwargs will look at object.__init__ and see that
                 # it takes *args and **kwargs and look for those.
-                kwargs: Dict[str, Any] = {}
+                kwargs: Dict[str, Any] = {}  # type: ignore[no-redef]
                 params.assert_empty(cls.__name__)
             else:
                 # This class has a constructor, so create kwargs for it.
                 constructor_to_inspect = cast(Callable[..., T], constructor_to_inspect)
-                kwargs = create_kwargs(constructor_to_inspect, cls, params, **extras)
+                kwargs = create_kwargs(constructor_to_inspect, cls, params, extras)
 
             return constructor_to_call(**kwargs)  # type: ignore
 
@@ -704,7 +812,7 @@ class FromParams(CustomDetHash):
 
         def replace_object_with_params(o: Any) -> Any:
             if isinstance(o, FromParams):
-                return o.to_params().as_dict()
+                return o.to_params().as_dict(quiet=True)
             elif isinstance(o, (list, tuple, set)):
                 return [replace_object_with_params(i) for i in o]
             elif isinstance(o, dict):
