@@ -1,16 +1,15 @@
 import copy
 import inspect
-from typing import Callable, Generic, TypeVar, Type, Union, Optional, Dict, Any
+from typing import Any, Callable, Dict, Generic, Optional, Type, TypeVar, Union
 
 from .params import Params
 
-
 T = TypeVar("T")
 
 
 class Lazy(Generic[T]):
     """
-    This class is for use when constructing objects using :class:`~tango.common.from_params.FromParams`,
+    This class is for use when constructing objects using :class:`~tango.common.FromParams`,
     when an argument to a constructor has a `sequential dependency` with another argument to the same
     constructor.
 
@@ -64,7 +63,9 @@ class Lazy(Generic[T]):
 
     @property
     def constructor(self) -> Callable[..., T]:
-        if inspect.isclass(self._constructor):
+        from tango.common.from_params import FromParams
+
+        if inspect.isclass(self._constructor) and issubclass(self._constructor, FromParams):
 
             def constructor_to_use(**kwargs):
                 return self._constructor.from_params(  # type: ignore[union-attr]
