@@ -10,7 +10,6 @@ from pathlib import Path
 from typing import Any, Dict, Iterable, List, Optional, Set, TypeVar, Union
 
 import yaml
-from overrides import overrides
 
 # _jsonnet doesn't work on Windows, so we have to use fakes.
 try:
@@ -33,6 +32,7 @@ except ImportError:
 
 from .aliases import PathOrStr
 from .exceptions import ConfigurationError
+from .util import could_be_class_name
 
 logger = logging.getLogger(__name__)
 
@@ -252,7 +252,6 @@ class Params(MutableMapping):
             self.params = _replace_none(params)
         self.history = history
 
-    @overrides
     def pop(self, key: str, default: Any = DEFAULT, keep_as_dict: bool = False) -> Any:
 
         """
@@ -274,7 +273,7 @@ class Params(MutableMapping):
         else:
             value = self.params.pop(key, default)
 
-        logger.info(f"{self.history}{key} = {value}")
+        logger.debug(f"{self.history}{key} = {value}")
         if keep_as_dict or _is_dict_free(value):
             return value
         else:
@@ -316,7 +315,6 @@ class Params(MutableMapping):
         else:
             raise ValueError("Cannot convert variable to bool: " + value)
 
-    @overrides
     def get(self, key: str, default: Any = DEFAULT):
         """
         Performs the functionality associated with ``dict.get(key)`` but also checks for returned
@@ -368,11 +366,11 @@ class Params(MutableMapping):
         """
         default = choices[0] if default_to_first_choice else self.DEFAULT
         value = self.pop(key, default)
-        ok_because_class_name = allow_class_names and "." in value
+        ok_because_class_name = allow_class_names and could_be_class_name(value)
         if value not in choices and not ok_because_class_name:
             key_str = self.history + key
             message = (
-                f"{value} not in acceptable choices for {key_str}: {choices}. "
+                f"'{value}' not in acceptable choices for {key_str}: {choices}. "
                 "You should either use the --include-package flag to make sure the correct module "
                 "is loaded, or use a fully qualified class name in your config file like "
                 """{"model": "my_module.models.MyModel"} to have it imported automatically."""
@@ -407,7 +405,7 @@ class Params(MutableMapping):
                     new_local_history = history + key + "."
                     log_recursively(value, new_local_history)
                 else:
-                    logger.info(f"{history}{key} = {value}")
+                    logger.debug(f"{history}{key} = {value}")
 
         log_recursively(self.params, self.history)
         return params_as_dict
