@@ -3,12 +3,12 @@ import os
 import shutil
 import tempfile
 from contextlib import contextmanager
-from copy import deepcopy
 from pathlib import Path
 from typing import Any, Dict, List, Optional, Union, cast
 
-from .aliases import PathOrStr
-from .registrable import Registrable
+from tango.common.aliases import PathOrStr
+from tango.common.logging import initialize_logging, teardown_logging
+from tango.common.params import Params
 
 
 class TangoTestCase:
@@ -54,32 +54,19 @@ class TangoTestCase:
         self.TEST_DIR = Path(tempfile.mkdtemp(prefix="tango_tests"))
         os.makedirs(self.TEST_DIR, exist_ok=True)
 
-    @classmethod
-    def setup_class(cls):
-        # During teardown we'll restore the state of `Registrable`'s internal registry
-        # to make sure any registered mock test classes are removed so they don't conflict
-        # with other tests.
-        cls._original_registry = deepcopy(Registrable._registry)
-
     def teardown_method(self):
         shutil.rmtree(self.TEST_DIR)
 
-    @classmethod
-    def teardown_class(cls):
-        Registrable._registry = cls._original_registry
-
     def run(
         self,
-        config: Union[PathOrStr, Dict[str, Any]],
+        config: Union[PathOrStr, Dict[str, Any], Params],
         overrides: Optional[Union[Dict[str, Any], str]] = None,
         include_package: Optional[List[str]] = None,
     ) -> Path:
         from tango.__main__ import TangoGlobalSettings, _run
 
-        from .params import Params
-
-        if isinstance(config, dict):
-            params = Params(config)
+        if isinstance(config, (dict, Params)):
+            params = config if isinstance(config, Params) else Params(config)
             config = self.TEST_DIR / "config.json"
             params.to_file(cast(Path, config))
 
@@ -88,29 +75,33 @@ class TangoTestCase:
 
             overrides = json.dumps(overrides)
 
-        run_dir = self.TEST_DIR / "run"
-        _run(
+        return _run(
             TangoGlobalSettings(),
             str(config),
-            directory=str(run_dir),
+            workspace_dir=str(self.TEST_DIR / "workspace"),
             overrides=overrides,
             include_package=include_package,
+            start_server=False,
         )
-        return run_dir
 
 
 @contextmanager
 def run_experiment(
-    config: Union[PathOrStr, Dict[str, Any]], overrides: Optional[Union[Dict[str, Any], str]] = None
+    config: Union[PathOrStr, Dict[str, Any], Params],
+    overrides: Optional[Union[Dict[str, Any], str]] = None,
+    file_friendly_logging: bool = True,
+    include_package: Optional[List[str]] = None,
 ):
     """
     A context manager to make testing experiments easier. On ``__enter__`` it runs
-    the experiment and returns the path to the cache directory, a temporary directory that will be
+    the experiment and returns the path to the run directory, a temporary directory that will be
     cleaned up on ``__exit__``.
     """
+    initialize_logging(enable_click_logs=True, file_friendly_logging=file_friendly_logging)
     test_case = TangoTestCase()
     try:
         test_case.setup_method()
-        yield test_case.run(config, overrides=overrides)
+        yield test_case.run(config, overrides=overrides, include_package=include_package)
     finally:
         test_case.teardown_method()
+        teardown_logging()
