@@ -81,12 +81,16 @@ You could then run this experiment with a config that looks like this:
     from tango.common.testing import run_experiment
     from tango.common.registrable import Registrable
 
-    # Don't cache results, otherwise we'll have a pickling error.
+    # Pickling the model fails because the class is defined ad hoc, not in a module.
+    # So we put in this hack to pickle a 0 instead of the Model.
+    def _return_zero(self):
+        return (int, (0,))
+    BasicRegression.__reduce__ = _return_zero
+
     with run_experiment(
-        "test_fixtures/integrations/torch/train.jsonnet",
-        overrides="{'steps.train.cache_results':false}"
+        "test_fixtures/integrations/torch/train.jsonnet"
     ) as run_dir:
-        assert (run_dir / "step_cache").is_dir()
+        assert (run_dir / "train").is_dir(), "Output for the 'train' step was not produced."
     # Restore state of registry.
     del Registrable._registry[Step]["generate_data"]
     del Registrable._registry[Model]["basic_regression"]
@@ -100,12 +104,15 @@ For example,
 would produce the following output:
 
 .. testoutput::
+    :options: +ELLIPSIS
 
-    ● Starting run for "data"...
-    ✓ Finished run for "data"
-    ● Starting run for "train"...
-    Loading best weights from state_worker0_best.pt
-    ✓ Finished run for "train"
+    Starting new run ...
+    ● Starting step "data" (needed by "train") ...
+    ✓ Finished step "data"
+    ● Starting step "train" ...
+    ✓ Finished step "train"
+    ✓ The output for "train" is in ...
+    Finished run ...
 
 Tips
 ----
@@ -130,6 +137,7 @@ needs to raise the :class:`StopEarly` exception.
 __all__ = [
     "TorchFormat",
     "TorchTrainStep",
+    "TorchEvalStep",
     "Optimizer",
     "LRScheduler",
     "Model",
@@ -138,12 +146,17 @@ __all__ = [
     "Sampler",
     "ConcatTensorDictsCollator",
     "TrainCallback",
+    "EvalCallback",
     "TrainConfig",
     "StopEarlyCallback",
     "StopEarly",
+    "TrainingEngine",
+    "TorchTrainingEngine",
 ]
 
 from .data import ConcatTensorDictsCollator, DataCollator, DataLoader, Sampler
+from .eval import TorchEvalStep
+from .eval_callback import EvalCallback
 from .exceptions import StopEarly
 from .format import TorchFormat
 from .model import Model
@@ -151,3 +164,4 @@ from .optim import LRScheduler, Optimizer
 from .train import TorchTrainStep
 from .train_callback import StopEarlyCallback, TrainCallback
 from .train_config import TrainConfig
+from .training_engine import TorchTrainingEngine, TrainingEngine
