@@ -1,6 +1,6 @@
-from dataclasses import dataclass
+from dataclasses import asdict, dataclass
 from pathlib import Path
-from typing import List, Optional
+from typing import Any, Dict, List, Optional
 
 import torch
 
@@ -11,6 +11,11 @@ class TrainConfig:
     Encapsulates the parameters of :class:`TorchTrainStep`.
     """
 
+    step_id: str
+    """
+    The unique ID of the current step.
+    """
+
     work_dir: Path
     """
     The working directory for the training run.
@@ -41,6 +46,13 @@ class TrainConfig:
     The number of steps to train for.
     """
 
+    train_epochs: Optional[int] = None
+    """
+    The number of epochs to train for.
+
+    You cannot specify `train_steps` and `train_epochs` at the same time.
+    """
+
     validation_steps: Optional[int] = None
     """
     The number of validation steps.
@@ -61,21 +73,11 @@ class TrainConfig:
     Controls the frequency of checkpoints.
     """
 
-    validate_every: int = 100
+    validate_every: Optional[int] = None
     """
     Controls the frequency of the validation loop.
     """
 
-    amp: bool = False
-    """
-    Whether automatic mixed precision is enabled.
-    """
-
-    max_grad_norm: Optional[float] = None
-    """
-    Max gradient norm. If set, gradients are clipped.
-    """
-
     is_distributed: bool = False
     """
     Whether or not the training job is distributed.
@@ -86,9 +88,14 @@ class TrainConfig:
     The devices used (for distributed jobs).
     """
 
+    distributed_address: str = "127.0.0.1"
+    """
+    The IP address of the main distributed process.
+    """
+
     distributed_port: str = "54761"
     """
-    The port of the distributed backend.
+    The port of the main distributed process.
     """
 
     val_metric_name: str = "loss"
@@ -101,7 +108,7 @@ class TrainConfig:
     Should be ``True`` when the validation metric being tracked should be minimized.
     """
 
-    aggregate_val_metric: bool = True
+    auto_aggregate_val_metric: bool = True
     """
     Controls automatic aggregation of validation metric.
     """
@@ -118,6 +125,8 @@ class TrainConfig:
 
     _worker_local_default_device: Optional[torch.device] = None
 
+    _device_type: Optional[str] = None  # either "cuda" or "cpu"
+
     @property
     def worker_local_default_device(self) -> torch.device:
         """
@@ -139,6 +148,17 @@ class TrainConfig:
             self._worker_local_default_device = device
             return device
 
+    @property
+    def device_type(self) -> str:
+        if self._device_type is None:
+            device_type = (
+                "cpu" if self.worker_local_default_device == torch.device("cpu") else "cuda"
+            )
+            self._device_type = device_type
+            return device_type
+        else:
+            return self._device_type
+
     @property
     def is_local_main_process(self) -> bool:
         """
@@ -151,7 +171,7 @@ class TrainConfig:
         """
         The path to the latest state checkpoint file.
         """
-        return self.work_dir / f"state_worker{self.worker_id}.pt"
+        return self.work_dir / "checkpoint_state_latest"
 
     @property
     def best_state_path(self) -> Path:
@@ -159,10 +179,14 @@ class TrainConfig:
         The path to the best state checkpoint file according to the validation metric or training
         loss (if no validation split is given).
         """
-        return self.work_dir / f"state_worker{self.worker_id}_best.pt"
+        return self.work_dir / "checkpoint_state_best"
 
     def state_path_for_step(self, step: int) -> Path:
-        return self.work_dir / f"state_worker{self.worker_id}_step{step + 1}.pt"
+        return self.work_dir / f"checkpoint_state_step{step + 1}"
+
+    @property
+    def final_weights_path(self) -> Path:
+        return self.work_dir / "weights.pt"
 
     def should_log_this_step(self, step: int) -> bool:
         assert self.train_steps is not None
@@ -172,12 +196,9 @@ class TrainConfig:
         assert self.train_steps is not None
         return ((step + 1) % self.checkpoint_every == 0) or step == self.train_steps - 1
 
-    def should_validate_this_step(self, step: int) -> bool:
-        assert self.train_steps is not None
-        return self.validation_split is not None and (
-            ((step + 1) % self.validate_every == 0) or step == self.train_steps - 1
-        )
-
     def should_log_this_val_step(self, val_step: int) -> bool:
         assert self.validation_steps is not None
         return val_step % self.log_every == 0 or val_step == self.validation_steps - 1
+
+    def as_dict(self) -> Dict[str, Any]:
+        return {k: v for k, v in asdict(self).items() if not k.startswith("_")}
