@@ -1,10 +1,11 @@
 import os
-import sys
+import warnings
 from typing import Any, Dict, List, Optional
 
-from overrides import overrides
+import torch
 
 from tango.integrations.torch.train_callback import TrainCallback
+from tango.integrations.torch.util import peak_gpu_memory
 
 
 @TrainCallback.register("wandb::log")
@@ -67,11 +68,17 @@ class WandbTrainCallback(TrainCallback):
         super().__init__(*args, **kwargs)
 
         if self.is_local_main_process and "WANDB_API_KEY" not in os.environ:
-            print(
+            warnings.warn(
                 "Missing environment variable 'WANDB_API_KEY' required to authenticate to Weights & Biases.",
-                file=sys.stderr,
+                UserWarning,
             )
 
+        _wandb_config = self.train_config.as_dict()
+        del _wandb_config["worker_id"]
+        _wandb_config["work_dir"] = str(self.train_config.work_dir.resolve())
+        if wandb_config is not None:
+            _wandb_config.update(wandb_config)
+
         self._watch_model = watch_model
         self._run_id: Optional[str] = None
         self._wandb_kwargs: Dict[str, Any] = dict(
@@ -79,15 +86,21 @@ class WandbTrainCallback(TrainCallback):
             project=project,
             entity=entity,
             group=group,
-            name=name,
+            name=name or self._get_run_name_from_step(),
             notes=notes,
-            config=wandb_config,
+            config=_wandb_config,
             tags=tags,
             anonymous="allow",
             **(wandb_kwargs or {}),
         )
 
-    @overrides
+    def _get_run_name_from_step(self) -> str:
+        """
+        Generate a run name for W&B based on the step unique ID.
+        """
+        step_name, step_hash = self.step_id.split("-", maxsplit=1)
+        return f"{step_name}-{step_hash[:6]}"
+
     def state_dict(self) -> Dict[str, Any]:
         if self.is_local_main_process:
             return {
@@ -96,14 +109,15 @@ class WandbTrainCallback(TrainCallback):
         else:
             return {}
 
-    @overrides
     def load_state_dict(self, state_dict: Dict[str, Any]) -> None:
         if self.is_local_main_process:
-            self._wandb_kwargs["resume"] = "auto"
+            self._wandb_kwargs["resume"] = "allow"
             self._run_id = state_dict["run_id"]
 
-    @overrides
     def pre_train_loop(self) -> None:
+        if torch.cuda.is_available():
+            torch.cuda.reset_peak_memory_stats()
+        peak_gpu_mbs = peak_gpu_memory()
         if self.is_local_main_process:
             import wandb
 
@@ -115,28 +129,40 @@ class WandbTrainCallback(TrainCallback):
             self.wandb.init(id=self._run_id, **self._wandb_kwargs)
 
             if self._watch_model:
-                self.wandb.watch(self.model)
+                self.wandb.watch(self.training_engine.model)
+            metrics = {f"sys/worker{rank}_peak_gpu_mem": mbs for rank, mbs in peak_gpu_mbs.items()}
+            metrics["epoch"] = 0
+            self.wandb.log(metrics, step=0)
 
-    @overrides
-    def post_train_loop(self) -> None:
+    def post_train_loop(self, step: int, epoch: int) -> None:
         if self.is_local_main_process:
             self.wandb.finish()
 
-    @overrides
-    def log_batch(self, step: int, batch_loss: float) -> None:
+    def log_batch(self, step: int, epoch: int, batch_loss: float) -> None:
+        peak_gpu_mbs = peak_gpu_memory()
         if self.is_local_main_process:
+            metrics = {
+                "train/loss": batch_loss,
+                "train/lr": self.training_engine.optimizer.param_groups[0]["lr"],
+                "epoch": epoch,
+            }
+            metrics.update(
+                {f"sys/worker{rank}_peak_gpu_mem": mbs for rank, mbs in peak_gpu_mbs.items()}
+            )
             self.wandb.log(
-                {"train/loss": batch_loss, "train/lr": self.optimizer.param_groups[0]["lr"]},
-                step=step,
+                metrics,
+                step=step + 1,
             )
 
-    @overrides
-    def post_val_loop(self, step: int, val_metric: float, best_val_metric: float) -> None:
+    def post_val_loop(
+        self, step: int, epoch: int, val_metric: float, best_val_metric: float
+    ) -> None:
         if self.is_local_main_process:
             self.wandb.log(
                 {
                     f"val/{self.train_config.val_metric_name}": val_metric,
                     f"val/best_{self.train_config.val_metric_name}": best_val_metric,
+                    "epoch": epoch,
                 },
-                step=step,
+                step=step + 1,
             )
